
package com.viavisolutions.cmp.signalworkshop;

import java.io.IOException;
import java.net.URI;

import javax.websocket.ClientEndpoint;
import javax.websocket.CloseReason;
import javax.websocket.ContainerProvider;
import javax.websocket.OnClose;
import javax.websocket.OnError;
import javax.websocket.OnMessage;
import javax.websocket.OnOpen;
import javax.websocket.Session;
import javax.websocket.WebSocketContainer;
import com.google.gson.Gson;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;


@ClientEndpoint
public class ApiClient {
	
	static final String clientId = "SVTJavaApiClient";
	static final String serverId = "SignalWorkshopService";
	
	static final String servicesReplyMsg =
			"{\n" +
    		"  \"MessageID\": [ 9, 1, 0, 4 ],\n" +
    		"  \"Data\": {\n" +
    		"    \"Services\": []\n" +
    		"  }\n" +
    		"}\n";
	
	/// This will subscribe to all messages from SignalWorkshop
	static final String subscriptionsReplyMsg =
			"{\n" +
			"  \"MessageID\": [ 9, 1, 0, 8], \n" +
			"  \"Data\": {\n" +
			"    \"Subscriptions\": [\n" +
			"      {\n" +
			"        \"MsgKey\": {\n" +
			"           \"MessageID\": [65535,65535,65535,65535],\n" +
			"           \"Sender\": [\"" + serverId + "\"],\n" +
			"           \"Destination\":[\"" + clientId + "\"]\n" +
			"        },\n" +
			"        \"Count\":1\n" +
			"      }\n" +
            "    ]" +
			"  }" +
			"}";
	
	static final String forwarderStartedMsg =
			"{" +
    		"  \"MessageID\": [ 9, 1, 0, 1 ]," +
    		"  \"Data\": { }" +
    		"}";
	
	static final String getServicesMsg =
			"{" +
    		"  \"MessageID\": [ 9, 1, 0, 3 ]," +
    		"  \"Data\": { }" +
    		"}";
	
	static final String getSubscriptionsMsg =
			"{" +
    		"  \"MessageID\": [ 9, 1, 0, 7 ]," +
    		"  \"Data\": { }" +
    		"}";
	
	static final String getCenterFreqMsg = 
		"{" +
        "  \"MessageID\": [ 4, 0, 0, 18]," +
        "  \"SenderID\": [ \"" + clientId + "\" ]," +
        "  \"DestID\": [\"" + serverId + "\", \"CenterFrequency\"]," +
        "  \"Data\": { }" +
        "}";
	
	static final String setCenterFreqMsg =
		"{" +
		"  \"MessageID\": [2,0,1,7]," +
		"  \"SenderID\":[\"" + clientId + "\"]," +
		"  \"DestID\":[\"" + serverId + "\"]," +
	    "  \"Data\":{" +
	    "    \"event\":1," +
	    "    \"param\":[{" +
	    "      \"MessageID\":[4,0,0,2]," +
	    "      \"Data\":{\"d\":<freq>}" +
	    "    }]" +
	    "  }" +
	    "}";
	
	
    Session userSession = null;
    boolean servicesSent = false;
    boolean subscriptionsSent = false;
    boolean clientSetup = false;
    
    double centerFreq = 0.0;
    Object centerFreqObj = new Object();
    
    static void debugPrint(String s)
    {
    	//System.out.println(s);
    }
    
    /**
     * @param endpointURI URI for client to connect to.  Example: ws://localhost:9024
     */
    public ApiClient(URI endpointURI)
    {
        try
        {
            WebSocketContainer container = ContainerProvider.getWebSocketContainer();
            userSession = container.connectToServer(this, endpointURI);
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
    }

    /**
     * Callback hook for Connection open events.
     *
     * @param session the session which is opened.
     */
    @OnOpen
    public void onOpen(Session session)
    {
    	debugPrint("opening websocket");
    }

    /**
     * Callback hook for Connection close events.
     *
     * @param session the session which is getting closed.
     * @param reason the reason for connection close
     */
    @OnClose
    public void onClose(Session session, CloseReason reason)
    {
    	debugPrint("closing websocket");
        this.userSession = null;
    }

    /**
     * Callback hook for Message Events. This method will be invoked when a client send a message.
     *
     * @param session the session which received the message
     * @param message The text message
     */
    @OnMessage
    public void onMessage(Session session, String message)
    {
    	debugPrint("Received message " + message);
    	
    	Gson gson = new Gson();
    	MsgHeader header = gson.fromJson(message, MsgHeader.class);
    	
        // is it a forwarder message?
        if(header.MessageID[0] == 9)
        {
        	try
        	{
	        	if(header.msgIdEqual(MsgHeader.FORWARDER_GET_SERVICES_ID))
	        	{
	        		debugPrint("Sending services msg");
	        		session.getBasicRemote().sendText(servicesReplyMsg);
	        		servicesSent = true;
	        		if(subscriptionsSent)
	        			clientSetup = true;
	        	}
	        	else if(header.msgIdEqual(MsgHeader.FORWARDER_GET_SUBSCRIPTIONS_ID))
	        	{
	        		debugPrint("Sending services msg");
	        		debugPrint(subscriptionsReplyMsg);
	        		session.getBasicRemote().sendText(subscriptionsReplyMsg);
	        		subscriptionsSent = true;
	        		if(servicesSent)
	        			clientSetup = true;
	        	}
	        	else if(header.msgIdEqual(MsgHeader.FORWARDER_STARTED_ID))
	        	{
	        		debugPrint("Sending forwarder started msgs");
	        		
	        		session.getBasicRemote().sendText(forwarderStartedMsg);
	        		session.getBasicRemote().sendText(getServicesMsg);
	        		session.getBasicRemote().sendText(getSubscriptionsMsg);
	        		
	        		debugPrint("Done sending start mesgs");
	        	}
        	}
        	catch(IOException e)
        	{
        		System.out.println("onMessage caught IOException " + e.getMessage());
        	}
        }
        else if(header.msgIdEqual(MsgHeader.DOUBLE_ATTR_ID))
        {
        	if(header.hasSenderId("CenterFrequency"))
        	{
	        	synchronized(centerFreqObj)
	        	{
		        	JsonObject jsonObj = new JsonParser().parse(message).getAsJsonObject();
		        	centerFreq = jsonObj.getAsJsonObject("Data").get("d").getAsDouble();
		        	centerFreqObj.notifyAll();
	        	}
        	}
        }
    }
    
    @OnError
    public void onError(Session session, Throwable thr)
    {
    	System.out.println("Client onError " + thr.getMessage());
    }

    /**
     * Send a message.
     *
     * @param message
     */
    public void sendMessage(String message)
    {
        userSession.getAsyncRemote().sendText(message);
    }
    
    /**
     * @return returns true once the setup messages for the client are complete.
     */
    public boolean isSetup()
    {
    	return clientSetup;
    }
    
    /**
     * @return Current center frequency from SignalWorkshop.  Frequency is in MHz
     */
    public double getCenterFreq()
    {
    	try
    	{
    		synchronized(centerFreqObj)
    		{
    			debugPrint("Sending " + getCenterFreqMsg);
	        	userSession.getAsyncRemote().sendText(getCenterFreqMsg);
	    		centerFreqObj.wait();
    		}
    	}
    	catch(InterruptedException e)
    	{
    		System.out.println("getCenterFreq wait interrupted " + e.getMessage());
    	}
    	
    	return centerFreq;
    }
    
    /**
     * Set center frequency in SignalWorkshop
     * 
     * @param freq Center Frequency in MHz to set in SignalWorkshop
     */
    public void setCenterFreq(double freq) throws IOException
    {
    	String json = setCenterFreqMsg.replace("<freq>", String.valueOf(freq));
    	debugPrint("Sending message:");
    	debugPrint(json);
    	userSession.getBasicRemote().sendText(json);
    }

    // If args[1] is provided, it is the ip address.  If it isn't provided then
    // use localhost
	public static void main(String args[])
	{
		ApiClient client;
		
		try
		{
			final String uri;
			if(args.length > 0)
				uri = "ws://" + args[0] + ":9024";
			else
				uri = "ws://localhost:9024";
			
			client = new ApiClient(new URI(uri));
			
			while(!client.isSetup())
			{
				Thread.sleep(100);
			}
			
			double freq = client.getCenterFreq();
			
			System.out.println("Received center frequency " + freq);
			System.out.println("Setting center frequency to " + (freq + 1.0));
			
			client.setCenterFreq(freq + 1.0);
		}
		catch(java.net.URISyntaxException e)
		{
			System.out.println("Error creating URI: " + e.getMessage());
		}
		catch(InterruptedException e)
		{
			System.out.println("Sleep interrupted!");
		}
		catch(IOException e)
		{
			System.out.println("IOException " + e.getMessage());
		}
	}

}
